--[[
Copyright (c) 2010 Pekka Karjalainen

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
]]

-- Handle creating the bulbs state and updating it.

-- The game state consists of the bulbs and count of bulbs currently on.
-- The game state is stored in a table of tables and may contain values
-- "on", "off" and "none". The "none" values represent a place where there
-- is no bulb and they also serve as sentinels. The table of tables is
-- treated as a 2-dimensional array. 
-- Because the newBulbs function add extra edge rows and columns to
-- the bulbs state, indexing this array is corrected in the public
-- toggle function.

-- Public functions:
-- newBulbs(tableOfString, randomFunction) -> makes a bulbs state
-- toggle(state, x, y) -> makes a move at position x, y
-- look(state, x, y) -> returns the value at the position x, y

local assert, table, unpack = assert, table, unpack
module(...)

-- private

-- Generate an iterator over the position y, x and its four neighbors.
local function neighbors(y, x)
	local vals = {{y, x}, {y-1, x}, {y+1, x}, {y, x-1}, {y, x+1}}
	local index = 0
	return function()
		index = index + 1
		if index <= 5 then return unpack(vals[index]) end
	end
end

-- Toggle the bulb at position x, y and all four nearest.
-- Toggling "none" positions has no effect on the state.
local function implToggle(state, x, y)
	assert(y > 1 and x > 1)
	assert(y < #state.bulbs and x < #state.bulbs[1])
	if state.bulbs[y][x] == "none" then return end
	for ny, nx in neighbors(y, x) do
		if state.bulbs[ny][nx] == "off" then
			state.count = state.count + 1
			state.bulbs[ny][nx] = "on"
		elseif state.bulbs[ny][nx] == "on" then
			state.count = state.count - 1
			state.bulbs[ny][nx] = "off"
		end
	end
end

-- Check how many bulbs would be turned on by making the move y, x.
-- This is implemented by actually toggling the position at y, x and
-- then undoing this by toggling it again.
--
-- This turned out to be unnecessarily complicated, but it works all right.
local function wouldChange(state, x, y)
	local old = state.count
	implToggle(state, x, y)
	local diff = state.count - old
	implToggle(state, x, y)
	return diff
end

-- Toggle bulbs randomly and ensure that more than atLeast
-- are on in the starting position.
local function initRandom(state, random, atLeast)
	local w, h = #state.bulbs[1], #state.bulbs
	for y = 2, h-1 do
		for x = 2, w-1 do
			if random(2) == 1 then
				implToggle(state, x, y)
			end
		end
	
	end
	-- Iteration cap of 1000 is much larger than any current level needs.
	local iter = 0
	while state.count < atLeast and iter < 1000 do
		local tx, ty = random(2, w-1), random(2, h-1)
		iter = iter + 1
		if wouldChange(state, tx, ty) > 0 then
			implToggle(state, tx, ty)
		end
	end
end
	

-- public

-- Correct the table indexing and call private toggle function.
function toggle(state, x, y)
	implToggle(state, x+1, y+1)
end

-- See what is at position x, y.
function look(state, x, y)
	return state.bulbs[y+1][x+1]
end

-- Create a new bulbs state from a table of strings.
-- It is assumed the strings form a rectangle. This function
-- surrounds the shape with sentinel "none" locations. Hence the
-- width and height of the final state is two larger than what is
-- derived from the strings.
--
-- The strings should consist of two values. # for "none" values and
-- . for bulbs.
function newBulbs(tos, random)
	local none, bulb = '#', '.'
	local width = #tos[1] + 2
	local height = #tos + 2 
	local state = {bulbs = {}, count = 0, height = height-2,
			width = width - 2}
	local bulbs = 0
	for y = 1, height do
		local line = {}
		-- top and bottom rows
		if y == 1 or y == height then
			for x = 1, width do
				table.insert(line, "none")
			end
		-- other rows
		else
			table.insert(line, "none")
			local row = tos[y-1]
			assert(#row == width - 2, "Inconsistent line widths")
			for x = 1, width-2 do
				local char = row:sub(x, x)
				if char == none then
					table.insert(line, "none")
				elseif char == bulb then
					table.insert(line, "off")
					bulbs = bulbs + 1
				else
					assert(false, "Illegal char in string")
				end
			end
			table.insert(line, "none")
		end
		table.insert(state.bulbs, line)
	end
	initRandom(state, random, bulbs/2)
	return state
end

