--[[
Copyright (c) 2010 Pekka Karjalainen

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
]]

-- The main program relies on just one module and a datafile that is read
-- and parsed at start. We use the Lua standard random generator, so
-- we seed it from the current time.

require 'bulbs'
math.randomseed(os.time())

local WIDTH, HEIGHT = 800, 600
local TEXTPOS
if love._version == 70 then TEXTPOS = 5 else TEXTPOS = 17 end
local BACKGROUND = {105, 135, 175}
local FOREGROUND = {155, 235, 155}
local MESSAGE_COLOR = {15, 15, 63}
local SELECTION_COLOR = {0, 0, 0, 63}
local LINE_WIDTH = 3

local MESSAGE_PLAYING = 
	"Clear the board by turning off all the lights. 'R' resets."
local MESSAGE_WON =
	"You have completed this level. Press space to proceed."
local MESSAGE_COMPLETED=
	"You have completed all the levels. Congratulations!"

-- Actual size of squares varies due to scaling. The value 40 was used
-- in testing, so we might as well keep it.
local SIZE = 40

local game = {}
local levels

-- Quick and dirty parsing. See levels.txt for explanation of the format
-- being parsed. Error handling is poor.
function parseLevels(filename)
	local comment = "@"
	local levels = {}
	local current = {}
	levelFile = love.filesystem.newFile(filename)
	levelFile:open("r")
	for line in levelFile:lines() do
		if #line == 0 or line:sub(1,1) == comment or
			line:sub(1,1) == " " then
			if #current > 0 then table.insert(levels, current) end
			current = {}
		else
			table.insert(current, line)
		end
	end
	levelFile:close()
	return levels
end

-- This functions sets the game to level passed as a param. Setting up
-- the same level again is a handy way to reset the board.
function setLevel(level)
	-- If the game is beat, we inform the player but can't proceed.
	if level > #levels then
		game.message = MESSAGE_COMPLETED
		return
	end
	local current = bulbs.newBulbs(levels[level], math.random)
	local w, h = current.width, current.height
	local sw, sh = WIDTH/(w+1)/SIZE, HEIGHT/(h+1)/SIZE
	game.scale = math.min(sw, sh)
	game.tx, game.ty = (WIDTH - (w+1)*game.scale*SIZE)/2, 
			   (HEIGHT - (h+1)*game.scale*SIZE)/2
	game.state = current
	game.won = false
	game.message = MESSAGE_PLAYING
end

-- We use just one hardcoded file for storing levels.
-- It didn't seem important to use a more robust solution.
function love.load()
	levels = parseLevels "levels.txt"
	game.level = 1
	setLevel(1)
	love.graphics.setBackgroundColor(unpack(BACKGROUND))
	love.graphics.setLineWidth(LINE_WIDTH)
end

-- The info line in the game always tells what keys can be pressed.
-- In addition, escape key quits game.
function love.keypressed(key)
	if key == 'escape' then love.event.push 'q'; return end
	if key == 'r' and not game.won then
		setLevel(game.level)
		return
	end 
	if key == ' ' and game.won then
		game.level = game.level + 1
		setLevel(game.level)
		return
	end
end

-- This game doesn't use the time diffs. The update function is all right
-- for sleeping a bit in case we don't have vsync active, finding which
-- tile (if any) the mouse is hovering over, and checking whether this
-- level has been just won.
function love.update()
	love.timer.sleep(10)
	local cx, cy
	local mx, my = love.mouse.getPosition()
	local tx, ty, scale = game.tx, game.ty, game.scale
	mx, my = (mx - tx) / scale, (my - ty) / scale
	cx, cy = math.floor((mx-0.625*SIZE)/SIZE) + 1,
		 math.floor((my-0.625*SIZE)/SIZE) + 1
	-- Make sure the position is actually within board limits.
	if cx < 1 or cx > game.state.width or
	   cy < 1 or cy > game.state.height then
		cx, cy = nil, nil
	end
	game.cx, game.cy = cx, cy
	if not game.won and game.state.count == 0 then
		game.won = true
		game.message = MESSAGE_WON
	end
end

-- We already know which board coordinates the mouse is over,
-- so we look at game.cx and game.cy directly.
function love.mousepressed(x, y, btn)
	if btn == 'l' and game.cx then
		bulbs.toggle(game.state, game.cx, game.cy)
	end
end

-- Draw the active area. Cx and cy have been checked to be over a legitimate
-- bulb, but we have to check the neighbors. This code is quite ugly, but
-- it works well enough.
function drawToggleArea(cx, cy)
	local nbors = {{cx-1, cy}, {cx+1, cy}, {cx, cy-1}, {cx, cy+1}}
	love.graphics.rectangle('fill', 0.625*SIZE+SIZE*(cx-1), 
					0.625*SIZE+SIZE*(cy-1), SIZE, SIZE)
	for _, pt in ipairs(nbors) do
		local cx, cy = pt[1], pt[2]
		if 1 <= cx and cx <= game.state.width and
		   1 <= cy and cy <= game.state.height and
		   bulbs.look(game.state, cx, cy) ~= "none" then
			love.graphics.rectangle('fill', 0.625*SIZE+SIZE*(cx-1),
				0.625*SIZE+SIZE*(cy-1), 40, 40)
		end
	end
end

-- We mostly draw rectangles. Better not get excited.
function love.draw()
	local w, h = game.state.width, game.state.height
	love.graphics.setColor(unpack(MESSAGE_COLOR))
	love.graphics.printf(game.message, 0, TEXTPOS, 800, "center")
	love.graphics.setColor(unpack(FOREGROUND))
	love.graphics.translate(game.tx, game.ty)
	love.graphics.scale(game.scale)
	for y = 1, h do
		for x = 1, w do
			local val = bulbs.look(game.state, x, y)
			local style
			if val == 'on' then style = 'fill' end
			if val == 'off' then style = 'line' end
			if style then
				love.graphics.rectangle(style, 
					0.75*SIZE+SIZE*(x-1),
					0.75*SIZE+SIZE*(y-1),
					0.75*SIZE, 0.75*SIZE)
			end
		end
	end
	if game.cx and bulbs.look(game.state, game.cx, game.cy) ~= 'none' then
		love.graphics.setColor(unpack(SELECTION_COLOR))
		drawToggleArea(game.cx, game.cy)
	end
end
